==> mule-module-troubleshooting-4.9.10-20250902.015123-6-sources.jar: org/mule/runtime/module/troubleshooting/internal/operations/EventDumpOperation.java <==
/*
 * Copyright 2023 Salesforce, Inc. All rights reserved.
 * The software in this package is published under the terms of the CPAL v1.0
 * license, a copy of which has been included with this distribution in the
 * LICENSE.txt file.
 */
package org.mule.runtime.module.troubleshooting.internal.operations;

import static java.lang.String.format;
import static java.lang.System.lineSeparator;

import static org.apache.commons.lang3.StringUtils.leftPad;
import static org.apache.commons.lang3.time.DurationFormatUtils.formatDuration;

import org.mule.runtime.core.api.context.notification.FlowCallStack;
import org.mule.runtime.core.api.context.notification.FlowStackElement;
import org.mule.runtime.core.api.event.EventContextService;
import org.mule.runtime.core.api.event.EventContextService.FlowStackEntry;
import org.mule.runtime.deployment.model.api.application.Application;
import org.mule.runtime.module.deployment.api.DeploymentService;
import org.mule.runtime.module.troubleshooting.api.ArgumentDefinition;
import org.mule.runtime.module.troubleshooting.api.TroubleshootingOperation;
import org.mule.runtime.module.troubleshooting.api.TroubleshootingOperationCallback;
import org.mule.runtime.module.troubleshooting.api.TroubleshootingOperationDefinition;
import org.mule.runtime.module.troubleshooting.internal.DefaultArgumentDefinition;
import org.mule.runtime.module.troubleshooting.internal.DefaultTroubleshootingOperationDefinition;

import java.io.IOException;
import java.io.Writer;
import java.util.Collection;
import java.util.List;
import java.util.Map.Entry;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * Operation used to collect an event dump.
 * <p>
 * The name of the operation is "events".
 * <p>
 * Arguments:
 * <ul>
 * <li>application (Optional): The application to collect the event dump from</li>
 * </ul>
 */
public class EventDumpOperation implements TroubleshootingOperation {

  public static final String EVENT_DUMP_OPERATION_NAME = "events";
  public static final String EVENT_DUMP_OPERATION_DESCRIPTION = "Collects an EventDump of currently active events";

  public static final String APPLICATION_ARGUMENT_NAME = "application";
  public static final String APPLICATION_ARGUMENT_DESCRIPTION = "Application to collect the event dump from";

  private static final TroubleshootingOperationDefinition definition = createOperationDefinition();

  private final DeploymentService deploymentService;

  public EventDumpOperation(DeploymentService deploymentService) {
    this.deploymentService = deploymentService;
  }

  @Override
  public TroubleshootingOperationDefinition getDefinition() {
    return definition;
  }

  @Override
  public TroubleshootingOperationCallback getCallback() {
    return (arguments, writer) -> {
      final String applicationName = arguments.get(APPLICATION_ARGUMENT_NAME);
      if (applicationName == null) {
        writeFlowStacksForAllApplications(writer);
      } else {
        Application application = deploymentService.findApplication(applicationName);
        writeFlowStackEntries(application, writer);
      }
    };
  }

  private static void writeFlowStacksFor(Application application, Writer writer)
      throws IOException {
    final var appsTitle = "Active Events for application '" + application.getArtifactName() + "'";
    writer.write(appsTitle + lineSeparator());
    writer.write(leftPad("", appsTitle.length(), "-") + lineSeparator());
    writer.write(lineSeparator());

    writeFlowStackEntries(application, writer);
  }

  private void writeFlowStacksForAllApplications(Writer writer) throws IOException {
    for (Application application : deploymentService.getApplications()) {
      writeFlowStacksFor(application, writer);
    }
  }

  private static void writeFlowStackEntries(Application application, Writer writer) throws IOException {
    EventContextService eventContextService = application
        .getArtifactContext()
        .getRegistry()
        .lookupByName(EventContextService.REGISTRY_KEY)
        .map(EventContextService.class::cast)
        .orElseThrow(() -> new IllegalArgumentException(format("Could not get EventContextService for application %s.",
                                                               application.getArtifactName())));

    final var currentlyActiveFlowStacks = eventContextService.getCurrentlyActiveFlowStacks();
    final var roots = hierarchicalEvents(currentlyActiveFlowStacks);

    writer.write(format(""
        + "Total Event Contexts: %6d%n"
        + "Total Root Contexts:  %6d%n"
        + "%n",
                        currentlyActiveFlowStacks.size(),
                        roots.size()));

    for (NestedEventsNode rootFlowStackNode : roots) {
      if (rootFlowStackNode.getDirectChidren().isEmpty()) {
        writer.write(formatFlowStack(rootFlowStackNode.getFlowStack()));
        writer.write(lineSeparator());
      } else {
        writer.write(format("\"%s\" hierarchy%n%n", rootFlowStackNode.getEventId()));
        writeHierarchy(writer, rootFlowStackNode, 0);
      }
    }
  }

  private static Collection<NestedEventsNode> hierarchicalEvents(final List<FlowStackEntry> currentlyActiveFlowStacks) {
    // use treesets to ensure consistent sorting in the report

    var nodesById = new TreeMap<String, NestedEventsNode>();
    for (FlowStackEntry fs : currentlyActiveFlowStacks) {
      final NestedEventsNode node = new NestedEventsNode(fs.getEventId(), fs);
      nodesById.put(fs.getEventId(), node);
    }

    var roots = new TreeMap<String, NestedEventsNode>();
    for (Entry<String, NestedEventsNode> node : nodesById.entrySet()) {
      final var parentEventId = node.getValue().getFlowStack().getParentEventId();
      if (parentEventId != null) {
        nodesById.get(parentEventId).addDirectChild(node.getValue());
      } else {
        roots.put(node.getKey(), node.getValue());
      }
    }
    return roots.values();
  }

  private static void writeHierarchy(Writer writer, NestedEventsNode node, int identIndex) throws IOException {
    for (NestedEventsNode childNode : node.getDirectChidren()) {
      writeHierarchy(writer, childNode, identIndex + 4);
    }

    writer.write(formatFlowStack(node.getFlowStack()).indent(identIndex));
    writer.write(lineSeparator());
  }

  private static class NestedEventsNode {

    private final String eventId;

    private final FlowStackEntry flowStack;

    private SortedMap<String, NestedEventsNode> directChildren = new TreeMap<>();

    public NestedEventsNode(String eventId, FlowStackEntry flowStack) {
      this.eventId = eventId;
      this.flowStack = flowStack;
    }

    public String getEventId() {
      return eventId;
    }

    public FlowStackEntry getFlowStack() {
      return flowStack;
    }

    public void addDirectChild(NestedEventsNode child) {
      directChildren.put(child.getEventId(), child);
    }

    public Collection<NestedEventsNode> getDirectChidren() {
      return directChildren.values();
    }

  }

  private static String formatFlowStack(FlowStackEntry fs) {
    return format("\"%s\", running for: %s, state: %s%n%s",
                  fs.getEventId(),
                  formatDuration(fs.getExecutingTime().toMillis(), "mm:ss.SSS"),
                  fs.getState().name(),
                  flowCallStackString(fs.getFlowCallStack()).indent(4));
  }

  // put this logic here so the current representation of stacks that is logged is not modified.
  private static String flowCallStackString(FlowCallStack flowCallStack) {
    StringBuilder stackString = new StringBuilder(256);

    int i = 0;
    final var flowStackElements = flowCallStack.getElements();
    for (FlowStackElement flowStackElement : flowStackElements) {
      stackString.append("at ").append(flowStackElement.toStringEventDumpFormat());
      if (++i != flowCallStack.getElements().size()) {
        stackString.append(lineSeparator());
      }
    }
    return stackString.toString();
  }

  private static TroubleshootingOperationDefinition createOperationDefinition() {
    return new DefaultTroubleshootingOperationDefinition(EVENT_DUMP_OPERATION_NAME, EVENT_DUMP_OPERATION_DESCRIPTION,
                                                         createApplicationArgumentDefinition());
  }

  private static ArgumentDefinition createApplicationArgumentDefinition() {
    return new DefaultArgumentDefinition(APPLICATION_ARGUMENT_NAME, APPLICATION_ARGUMENT_DESCRIPTION, false);
  }
}

==> mule-module-troubleshooting-4.9.12-20251126.092512-17-sources.jar: org/mule/runtime/module/troubleshooting/internal/operations/EventDumpOperation.java <==
/*
 * Copyright 2023 Salesforce, Inc. All rights reserved.
 * The software in this package is published under the terms of the CPAL v1.0
 * license, a copy of which has been included with this distribution in the
 * LICENSE.txt file.
 */
package org.mule.runtime.module.troubleshooting.internal.operations;

import static java.lang.String.format;
import static java.lang.System.lineSeparator;

import static org.apache.commons.lang3.StringUtils.leftPad;
import static org.apache.commons.lang3.time.DurationFormatUtils.formatDuration;

import org.mule.runtime.core.api.context.notification.FlowCallStack;
import org.mule.runtime.core.api.context.notification.FlowStackElement;
import org.mule.runtime.core.api.event.EventContextService;
import org.mule.runtime.core.api.event.EventContextService.FlowStackEntry;
import org.mule.runtime.deployment.model.api.application.Application;
import org.mule.runtime.module.deployment.api.DeploymentService;
import org.mule.runtime.module.troubleshooting.api.ArgumentDefinition;
import org.mule.runtime.module.troubleshooting.api.TroubleshootingOperation;
import org.mule.runtime.module.troubleshooting.api.TroubleshootingOperationCallback;
import org.mule.runtime.module.troubleshooting.api.TroubleshootingOperationDefinition;
import org.mule.runtime.module.troubleshooting.internal.DefaultArgumentDefinition;
import org.mule.runtime.module.troubleshooting.internal.DefaultTroubleshootingOperationDefinition;

import java.io.IOException;
import java.io.Writer;
import java.util.Collection;
import java.util.List;
import java.util.Map.Entry;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * Operation used to collect an event dump.
 * <p>
 * The name of the operation is "events".
 * <p>
 * Arguments:
 * <ul>
 * <li>application (Optional): The application to collect the event dump from</li>
 * </ul>
 */
public class EventDumpOperation implements TroubleshootingOperation {

  public static final String EVENT_DUMP_OPERATION_NAME = "events";
  public static final String EVENT_DUMP_OPERATION_DESCRIPTION = "Collects an EventDump of currently active events";

  public static final String APPLICATION_ARGUMENT_NAME = "application";
  public static final String APPLICATION_ARGUMENT_DESCRIPTION = "Application to collect the event dump from";

  private static final TroubleshootingOperationDefinition definition = createOperationDefinition();

  private final DeploymentService deploymentService;

  public EventDumpOperation(DeploymentService deploymentService) {
    this.deploymentService = deploymentService;
  }

  @Override
  public TroubleshootingOperationDefinition getDefinition() {
    return definition;
  }

  @Override
  public TroubleshootingOperationCallback getCallback() {
    return (arguments, writer) -> {
      final String applicationName = arguments.get(APPLICATION_ARGUMENT_NAME);
      if (applicationName == null) {
        writeFlowStacksForAllApplications(writer);
      } else {
        Application application = deploymentService.findApplication(applicationName);
        writeFlowStackEntries(application, writer);
      }
    };
  }

  private static void writeFlowStacksFor(Application application, Writer writer)
      throws IOException {
    final var appsTitle = "Active Events for application '" + application.getArtifactName() + "'";
    writer.write(appsTitle + lineSeparator());
    writer.write(leftPad("", appsTitle.length(), "-") + lineSeparator());
    writer.write(lineSeparator());

    writeFlowStackEntries(application, writer);
  }

  private void writeFlowStacksForAllApplications(Writer writer) throws IOException {
    for (Application application : deploymentService.getApplications()) {
      writeFlowStacksFor(application, writer);
    }
  }

  private static void writeFlowStackEntries(Application application, Writer writer) throws IOException {
    EventContextService eventContextService = application
        .getArtifactContext()
        .getRegistry()
        .lookupByName(EventContextService.REGISTRY_KEY)
        .map(EventContextService.class::cast)
        .orElseThrow(() -> new IllegalArgumentException(format("Could not get EventContextService for application %s.",
                                                               application.getArtifactName())));

    final var currentlyActiveFlowStacks = eventContextService.getCurrentlyActiveFlowStacks();
    final var roots = hierarchicalEvents(currentlyActiveFlowStacks);

    writer.write(format(""
        + "Total Event Contexts: %6d%n"
        + "Total Root Contexts:  %6d%n"
        + "%n",
                        currentlyActiveFlowStacks.size(),
                        roots.size()));

    for (NestedEventsNode rootFlowStackNode : roots) {
      if (rootFlowStackNode.getDirectChidren().isEmpty()) {
        writer.write(formatFlowStack(rootFlowStackNode.getFlowStack()));
        writer.write(lineSeparator());
      } else {
        writer.write(lineSeparator());
        writer.write(format("\"%s\" hierarchy%n%n", rootFlowStackNode.getEventId()));
        writeHierarchy(writer, rootFlowStackNode, 0);
      }
    }
  }

  private static Collection<NestedEventsNode> hierarchicalEvents(final List<FlowStackEntry> currentlyActiveFlowStacks) {
    // use treesets to ensure consistent sorting in the report

    var nodesById = new TreeMap<String, NestedEventsNode>();
    for (FlowStackEntry fs : currentlyActiveFlowStacks) {
      final NestedEventsNode node = new NestedEventsNode(fs.getEventId(), fs);
      nodesById.put(fs.getEventId(), node);
    }

    var roots = new TreeMap<String, NestedEventsNode>();
    for (Entry<String, NestedEventsNode> node : nodesById.entrySet()) {
      final var parentEventId = node.getValue().getFlowStack().getParentEventId();
      if (parentEventId != null) {
        nodesById.get(parentEventId).addDirectChild(node.getValue());
      } else {
        roots.put(node.getKey(), node.getValue());
      }
    }
    return roots.values();
  }

  private static void writeHierarchy(Writer writer, NestedEventsNode node, int identIndex, FlowCallStack parentStack)
      throws IOException {
    FlowCallStack currentStack = node.getFlowStack().getFlowCallStack();

    for (NestedEventsNode childNode : node.getDirectChidren()) {
      writeHierarchy(writer, childNode, identIndex + 4, currentStack);
    }

    writer.write(formatFlowStack(node.getFlowStack(), parentStack).indent(identIndex));
    writer.write(lineSeparator());
  }

  private static void writeHierarchy(Writer writer, NestedEventsNode node, int identIndex) throws IOException {
    writeHierarchy(writer, node, identIndex, null);
  }

  private static class NestedEventsNode {

    private final String eventId;

    private final FlowStackEntry flowStack;

    private SortedMap<String, NestedEventsNode> directChildren = new TreeMap<>();

    public NestedEventsNode(String eventId, FlowStackEntry flowStack) {
      this.eventId = eventId;
      this.flowStack = flowStack;
    }

    public String getEventId() {
      return eventId;
    }

    public FlowStackEntry getFlowStack() {
      return flowStack;
    }

    public void addDirectChild(NestedEventsNode child) {
      directChildren.put(child.getEventId(), child);
    }

    public Collection<NestedEventsNode> getDirectChidren() {
      return directChildren.values();
    }

  }

  private static String formatFlowStack(FlowStackEntry fs) {
    return formatFlowStack(fs, null);
  }

  private static String formatFlowStack(FlowStackEntry fs, FlowCallStack parentStack) {
    FlowCallStack currentStack = fs.getFlowCallStack();
    String stackString = parentStack != null
        ? flowCallStackStringUnique(currentStack, parentStack)
        : flowCallStackString(currentStack);

    return format("\"%s\", running for: %s, state: %s%n%s",
                  fs.getEventId(),
                  formatDuration(fs.getExecutingTime().toMillis(), "mm:ss.SSS"),
                  fs.getState().name(),
                  stackString.indent(4));
  }

  // put this logic here so the current representation of stacks that is logged is not modified.
  private static String flowCallStackString(FlowCallStack flowCallStack) {
    return flowCallStackString(flowCallStack.getElements());
  }

  private static String flowCallStackString(List<FlowStackElement> flowStackElements) {
    StringBuilder stackString = new StringBuilder(256);

    int i = 0;
    for (FlowStackElement flowStackElement : flowStackElements) {
      stackString.append("at ").append(flowStackElement.toStringEventDumpFormat());
      if (++i != flowStackElements.size()) {
        stackString.append(lineSeparator());
      }
    }
    return stackString.toString();
  }

  /**
   * Returns only the unique prefix of the child stack, excluding the common suffix that is already shown in the parent. This
   * helps reduce the size of the event dump by avoiding redundant information.
   * <p>
   * Since child contexts are created by cloning the parent's stack and then pushing new elements, the child stack always contains
   * all parent elements at the end. We can simply take the first (childSize - parentSize) elements as the unique prefix.
   */
  private static String flowCallStackStringUnique(FlowCallStack childStack, FlowCallStack parentStack) {
    List<FlowStackElement> childElements = childStack.getElements();
    List<FlowStackElement> parentElements = parentStack.getElements();

    if (childElements.isEmpty()) {
      return "";
    }

    if (parentElements.isEmpty()) {
      return flowCallStackString(childStack);
    }

    // Child stack is created by cloning parent and pushing new elements, so the last parentSize elements
    // are from the parent. Show only the unique prefix (first childSize - parentSize elements).
    int uniquePrefixLength = childElements.size() - parentElements.size();

    if (uniquePrefixLength <= 0) {
      // Child has same or fewer elements than parent, show only first element as continuation indicator
      return "at " + childElements.get(0).toStringEventDumpFormat();
    }

    return flowCallStackString(childElements.subList(0, uniquePrefixLength));
  }

  private static TroubleshootingOperationDefinition createOperationDefinition() {
    return new DefaultTroubleshootingOperationDefinition(EVENT_DUMP_OPERATION_NAME, EVENT_DUMP_OPERATION_DESCRIPTION,
                                                         createApplicationArgumentDefinition());
  }

  private static ArgumentDefinition createApplicationArgumentDefinition() {
    return new DefaultArgumentDefinition(APPLICATION_ARGUMENT_NAME, APPLICATION_ARGUMENT_DESCRIPTION, false);
  }
}
